###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import numpy as np
import pandas as pd
import os
from sciutil import SciUtil, SciException
import matplotlib.pyplot as plt
from scipy.stats import combine_pvalues

class SciRCMException(SciException):
    def __init__(self, message=''):
        Exception.__init__(self, message)

"""
Sci-RCM is the logical regulatory clustering of genes based on DNA-methylation, RNA-seq and Proteomics data.

| Methylation      | RNAseq    | Proteomics | Regulation driver_1          | Regulation driver_2     | Regulation_Grouping1 | Regulation_Grouping2 | Regulation_Grouping3 |
|------------------|-----------|------------|------------------------------|-------------------------|----------------------|----------------------|----------------------|
| Hypermethylation | DOWN      | DOWN       | Methylation increase (MDS)   | None                    | MDS                  | MDS                  | MDS                  |
| Hypermethylation | UP        | DOWN       | mRNA increase (TPDE)         | Protein decrease (TMDS) | TPDE+TMDS            | TPDE+TMDS            | TMDS                 |
| Hypermethylation | UP        | UP         | mRNA increase (TPDE)         | None                    | TPDE                 | TPDE                 | TPDE                 |
| Hypermethylation | DOWN      | UP         | Methylation increase (MDS)   | Protein increase (TMDE) | MDS+TMDE             | TMDE                 | TMDE                 |
| Hypermethylation | No Change | UP         | mRNA increase (TPDE)         | Protein increase (TMDE) | TPDE+TMDE            | TMDE                 | TMDE                 |
| Hypermethylation | No Change | DOWN       | mRNA increase (TPDE)         | Protein decrease (TMDS) | TPDE+TMDS            | TMDS                 | TMDS                 |
| Hypermethylation | UP        | No Change  | mRNA increase (TPDE)         | Protein decrease (TMDS) | TPDE+TMDS            | TPDE+TMDS            | TMDS                 |
| Hypermethylation | DOWN      | No Change  | Methylation increase (MDS)   | Protein increase (TMDE) | MDS+TMDE             | MDS+TMDE             | TMDE                 |
| Hypermethylation | No Change | No Change  | Methylation increase (ncRNA) | None                    | MDS-ncRNA            | MDS_ncRNA            | MDS_ncRNA            |
| Hypomethylation  | DOWN      | DOWN       | mRNA decrease (TPDS)         | None                    | TPDS                 | TPDS                 | TPDS                 |
| Hypomethylation  | UP        | DOWN       | Methylation decrease (MDE)   | Protein decrease (TMDS) | MDE+TMDS             | TMDS                 | TMDS                 |
| Hypomethylation  | UP        | UP         | Methylation decrease (MDE)   | None                    | MDE                  | MDE                  | MDE                  |
| Hypomethylation  | DOWN      | UP         | mRNA decrease (TPDS)         | Protein increase (TMDE) | TPDS+TMDE            | TPDS+TMDE            | TMDE                 |
| Hypomethylation  | No Change | UP         | mRNA decrease (TPDS)         | Protein increase (TMDE) | TPDS+TMDE            | TMDE                 | TMDE                 |
| Hypomethylation  | No Change | DOWN       | mRNA decrease (TPDS)         | Protein decrease (TMDS) | TPDS+TMDS            | TMDS                 | TMDS                 |
| Hypomethylation  | UP        | No Change  | Methylation decrease (MDE)   | Protein decrease (TMDS) | MDE+TMDS             | MDE+TMDS             | TMDS                 |
| Hypomethylation  | DOWN      | No Change  | mRNA decrease (TPDS)         | Protein increase (TMDE) | TPDS+TMDE            | TPDS+TMDE            | TMDE                 |
| Hypomethylation  | No Change | No Change  | Methylation decrease (ncRNA) | None                    | MDE+ncRNA            | MDE_ncRNA            | MDE_ncRNA            |
| No Change        | DOWN      | UP         | mRNA decrease (TPDS)         | Protein increase (TMDE) | TMDE                 | TPDS+TMDE            | TMDE                 |
| No Change        | UP        | DOWN       | mRNA increase (TPDE)         | Protein decrease (TMDS) | TMDS                 | TPDE+TMDS            | TMDS                 |
| No Change        | DOWN      | DOWN       | mRNA decrease (TPDS)         | None                    | TPDS                 | TPDS                 | TPDS                 |
| No Change        | UP        | UP         | mRNA increase (TPDE)         | None                    | TPDE                 | TPDE                 | TPDE                 |
| No Change        | No Change | UP         | Protein increase (TMDE)      | None                    | TMDE                 | TMDE                 | TMDE                 |
| No Change        | No Change | DOWN       | Protein decrease (TMDS)      | None                    | TMDS                 | TMDS                 | TMDS                 |
| No Change        | UP        | No Change  | mRNA increase (TPDE)         | Protein decrease (TMDS) | TPDE+TMDS            | TPDE+TMDS            | TMDS                 |
| No Change        | DOWN      | No Change  | mRNA decrease (TPDS)         | Protein increase (TMDE) | TPDS+TMDE            | TPDS+TMDE            | TMDE                 |
| No Change        | No Change | No Change  | NoChange                     | NoChange                | NoChange             | NoChange             | NoChange             |

Sci-RCM also interfaces with scivae (the variational autoencoder component that we use to generate an integrated rank).
"""


class SciRCM:

    def __init__(self, meth_file: str,
                 rna_file: str,
                 proteomics_file: str,
                 rna_logfc: str,
                 rna_padj: str,
                 meth_diff: str,
                 meth_padj: str,
                 prot_logfc: str,
                 prot_padj: str,
                 gene_id: str,
                 rna_padj_cutoff=0.05,
                 prot_padj_cutoff=0.05,
                 meth_padj_cutoff=0.05,
                 rna_logfc_cutoff=1.0,
                 prot_logfc_cutoff=0.5,
                 meth_diff_cutoff=10,
                 output_dir='.',
                 output_filename=None,
                 non_coding_genes=None,
                 debug_on=False, sep=',',
                 bg_type='P|(M&R)',
                 sciutil=None,
                 logfile=None,
                 reg_grp_1_lbl='Regulation_Grouping_1',
                 reg_grp_2_lbl='Regulation_Grouping_2',
                 reg_grp_3_lbl='Regulation_Grouping_3',
                 main_reg_label='Regulation_Grouping_2'):
        self.u = SciUtil() if sciutil is None else sciutil
        plt.rcParams['svg.fonttype'] = 'none'
        self.meth_diff = meth_diff
        self.merged_df = None
        self.prot_logfc = prot_logfc
        self.rna_logfc = rna_logfc
        self.rna_padj = rna_padj
        self.meth_padj = meth_padj
        self.prot_padj = prot_padj
        self.gene_id = gene_id
        self.logfile = open(logfile, "w+") if logfile is not None else None  # File for logging results
        self.rna_padj_cutoff, self.meth_padj_cutoff, self.prot_padj_cutoff = rna_padj_cutoff, meth_padj_cutoff, prot_padj_cutoff
        self.rna_logfc_cutoff, self.meth_diff_cutoff, self.prot_logfc_cutoff = rna_logfc_cutoff, meth_diff_cutoff, prot_logfc_cutoff
        self.debug = debug_on
        self.output_dir = output_dir
        self.reg_grp_1_lbl = reg_grp_1_lbl
        self.reg_grp_2_lbl = reg_grp_2_lbl
        self.reg_grp_3_lbl = reg_grp_3_lbl
        self.main_reg_label = main_reg_label
        if main_reg_label != reg_grp_2_lbl and main_reg_label != reg_grp_1_lbl and main_reg_label != reg_grp_3_lbl:
            self.u.err_p([f'ERROR: your main regulatory label (main_reg_label) must be one of: '
                          f'{reg_grp_1_lbl}, {reg_grp_2_lbl} or {reg_grp_3_lbl}, you passed: ', main_reg_label])
        self.bg_type = bg_type
        # Otherwise this will be in the specific ones.
        if prot_logfc and rna_logfc and meth_diff:
            self.bg_list = ['(P&M)|(P&R)', '(P&M)|(P&R)|(M&R)', '*', 'P&M&R', 'P|M|R', 'P|(M&R)']
            if bg_type not in self.bg_list:
                self.u.err_p(['ERROR: selected background type was not allowed, please choose from one of: ', self.bg_list,
                              '\n Note: | means OR and & means AND'])
            self.output_filename = output_filename if output_filename else f'scircm_r{rna_logfc_cutoff}-{rna_padj_cutoff}' \
                                                                       f'_p{prot_logfc_cutoff}-{prot_padj_cutoff}' \
                                                                       f'_m{meth_diff_cutoff}-{meth_padj_cutoff}.csv'
            self.meth_df = pd.read_csv(meth_file, sep=sep)
            self.rna_df = pd.read_csv(rna_file, sep=sep)
            self.prot_df = pd.read_csv(proteomics_file, sep=sep)

        self.non_coding_genes = non_coding_genes
        # Contains genes for the non-coding region (use for human only).
        self.df = None
        # Contains the vae data
        self.vae = None

    def run(self):
        # Merge the dataframes together
        self.merge_dfs()
        # Generate background
        self.df = self.gen_bg()
        # Calculate groups
        self.run_rcm()
        # Save the DF and return the groupings
        self.df.to_csv(os.path.join(self.output_dir, self.output_filename), index=False)
        return self.df

    @staticmethod
    def _get_bg_filter(bg_type, prot_val, rna_val, meth_val, prot_padj_cutoff, meth_padj_cutoff, rna_padj_cutoff):
        """
        Filters the data based on a background selection. We have defined the filter such that it is based on being
        at minimum "significant". The gene doesn't have to ALSO meet the threshold (i.e. being above or less than
        a certain logFC).
        """
        c = 0
        if bg_type == 'P|(M&R)':  # Protein AND (DNA methylation OR RNA)
            c = 1 if prot_val <= prot_padj_cutoff else 0
            c += 1 if rna_val <= rna_padj_cutoff and meth_val <= meth_padj_cutoff else 0
        elif bg_type == 'P|M|R':  # Protein OR methylation OR RNA
            c = 1 if prot_val <= prot_padj_cutoff else 0
            c += 1 if rna_val <= rna_padj_cutoff else 0
            c += 1 if meth_val <= meth_padj_cutoff else 0
        elif bg_type == 'P&M&R':  # Protein AND Methylation AND RNA
            c = 1 if (prot_val <= prot_padj_cutoff and rna_val <= rna_padj_cutoff and meth_val <= meth_padj_cutoff) else 0
        elif bg_type == '(P&M)|(P&R)|(M&R)':  # At least two are significant
            c = 1 if (prot_val <= prot_padj_cutoff and rna_val <= rna_padj_cutoff) else 0
            c += 1 if (prot_val <= prot_padj_cutoff and meth_val <= meth_padj_cutoff) else 0
            c += 1 if (rna_val <= rna_padj_cutoff and meth_val <= meth_padj_cutoff) else 0
        elif bg_type == '(P&M)|(P&R)':  # Protein and one other
            c = 1 if (prot_val <= prot_padj_cutoff and rna_val <= rna_padj_cutoff) else 0
            c += 1 if (prot_val <= prot_padj_cutoff and meth_val <= meth_padj_cutoff) else 0
        elif bg_type == '*':  # Use all genes as the background
            c = 1
        return c

    def gen_bg(self, bg_type=None, prot_padj_cutoff=None, rna_padj_cutoff=None, meth_padj_cutoff=None):
        """
        Generate a background dataset i.e. since the RCM requires at least two of the 3 datasets to
        have a p value beneath a threshold we reduce our dataset to be smaller.
        """
        bg_type = bg_type if bg_type else self.bg_type
        prot_padj_cutoff = prot_padj_cutoff if prot_padj_cutoff else self.prot_padj_cutoff
        meth_padj_cutoff = meth_padj_cutoff if meth_padj_cutoff else self.meth_padj_cutoff
        rna_padj_cutoff = rna_padj_cutoff if rna_padj_cutoff else self.rna_padj_cutoff
        filter_vals = np.zeros(len(self.merged_df))
        meth_padj_values = self.merged_df[self.meth_padj].values
        prot_padj_values = self.merged_df[self.prot_padj].values

        # Choose the background dataframe
        for i, rna_padj in enumerate(self.merged_df[self.rna_padj].values):
            c = self._get_bg_filter(bg_type, prot_val=prot_padj_values[i], rna_val=rna_padj,
                                    meth_val=meth_padj_values[i],
                                    prot_padj_cutoff=prot_padj_cutoff, meth_padj_cutoff=meth_padj_cutoff,
                                    rna_padj_cutoff=rna_padj_cutoff)
            filter_vals[i] = c

        df = self.merged_df.copy()
        # Filter the df to become the background df
        df['Number of significant datasets'] = filter_vals
        # Filter DF to only include those that are sig in two.
        df = df[df['Number of significant datasets'] >= 1]
        df = df.reset_index()  # Reset the index of the dataframe

        return df

    def merge_dfs(self):
        """
        Merge the supplied RNAseq, DNA methylation and proteomics dataframes together. We do an outer join. This can
        result in gene identifiers becoming disjoint (i.e. if there are two ensembl ID columns, we get two now).
        """
        self.rna_df = self.rna_df.set_index(self.gene_id)
        self.df = self.rna_df.merge(self.meth_df, on=self.gene_id, how='outer', suffixes=['_r', '_m'])
        self.df = self.df.merge(self.prot_df, on=self.gene_id, how='outer', suffixes=['', '_p'])
        # Fill any of the values with NAs (p values need to be filled with 1's)
        self.df[[self.meth_padj]] = self.df[[self.meth_padj]].fillna(value=1.0)
        self.df[[self.rna_padj]] = self.df[[self.rna_padj]].fillna(value=1.0)
        self.df[[self.prot_padj]] = self.df[[self.prot_padj]].fillna(value=1.0)
        # Fill the rest of the values with 0's
        self.df = self.df.fillna(0)  # Fill any other values with 0 i.e. any of the logFCs or statistics.
        self.df.to_csv(os.path.join(self.output_dir, "SIRCLE_merged_df.csv"), index=False)
        self.merged_df = self.df.copy()

    def run_rcm(self):
        """
        Runs the regulatory clustering and assigns labels for the three levels of clustering:
        1) Level 1: which is the most granular
        2) Level 2: produces a medium level of filtering (this is the default)
        3) Level 3: produces a general idea of what the regulational level is on the basis of the protein.
        :return:
        """
        lbls = ['None'] * len(self.df)
        # Initialise the regulatory labels
        self.df[self.reg_grp_1_lbl] = lbls
        self.df[self.reg_grp_2_lbl] = lbls
        self.df[self.reg_grp_3_lbl] = lbls

        if self.logfile is not None:
            self.logfile.write(f'MethylationState,RNAseqState,ProteinState,Label,NumGenes\n')

        # -------------- Clusters with hyper-methylation
        # Hypermethylation | DOWN      | DOWN       | Methylation increase (MDS)   | None
        self.get_grp(meth_c='pos', rna_c='neg', prot_c='neg', grp_id='MDS', reg_grp_1='MDS', reg_grp_3='MDS')

        # Hypermethylation | UP        | DOWN       | mRNA increase (TPDE)         | Protein decrease (TMDS)
        self.get_grp(meth_c='pos', rna_c='pos', prot_c='neg', grp_id='TPDE_TMDS', reg_grp_1='TPDE_TMDS', reg_grp_3='TMDS')

        # Hypermethylation | UP        | UP         | mRNA increase (TPDE)         | None
        self.get_grp(meth_c='pos', rna_c='pos', prot_c='pos', grp_id='TPDE', reg_grp_1='TPDE', reg_grp_3='TPDE')

        # Hypermethylation | DOWN      | UP         | Methylation increase (MDS)   | Protein increase (TMDE)
        self.get_grp(meth_c='pos', rna_c='neg', prot_c='pos', grp_id='TMDE', reg_grp_1='MDS_TMDE', reg_grp_3='TMDE')

        # Hypermethylation | No Change | UP         | mRNA increase (TPDE)         | Protein increase (TMDE)
        self.get_grp(meth_c='pos', rna_c='-', prot_c='pos', grp_id='TMDE', reg_grp_1='TPDE_TMDE', reg_grp_3='TMDE')

        # Hypermethylation | No Change | DOWN       | mRNA increase (TPDE)         | Protein decrease (TMDS)
        self.get_grp(meth_c='pos', rna_c='-', prot_c='neg', grp_id='TMDS', reg_grp_1='TPDE_TMDS', reg_grp_3='TMDS')

        # Hypermethylation | UP        | No Change  | mRNA increase (TPDE)         | Protein decrease (TMDS)
        self.get_grp(meth_c='pos', rna_c='pos', prot_c='-', grp_id='TPDE_TMDS', reg_grp_1='TPDE_TMDS', reg_grp_3='TMDS')

        # Hypermethylation | DOWN      | No Change  | Methylation increase (MDS)   | Protein increase (TMDE)
        self.get_grp(meth_c='pos', rna_c='neg', prot_c='-', grp_id='MDS_TMDE', reg_grp_1='MDS_TMDE', reg_grp_3='TMDE')

        # -------------- Clusters with hypo-methylation
        # Hypomethylation  | DOWN      | DOWN       | mRNA decrease (TPDS)         | None
        self.get_grp(meth_c='neg', rna_c='neg', prot_c='neg', grp_id='TPDS', reg_grp_1='TPDS', reg_grp_3='TPDS')

        # Hypomethylation  | UP        | DOWN       | Methylation decrease (MDE)   | Protein decrease (TMDS)
        self.get_grp(meth_c='neg', rna_c='pos', prot_c='neg', grp_id='TMDS', reg_grp_1='MDE_TMDS', reg_grp_3='TMDS')

        # Hypomethylation  | UP        | UP         | Methylation decrease (MDE)   | None
        self.get_grp(meth_c='neg', rna_c='pos', prot_c='pos', grp_id='MDE', reg_grp_1='MDE', reg_grp_3='MDE')

        # Hypomethylation  | DOWN      | UP         | mRNA decrease (TPDS)         | Protein increase (TMDE)
        self.get_grp(meth_c='neg', rna_c='neg', prot_c='pos', grp_id='TPDS_TMDE', reg_grp_1='TPDS_TMDE', reg_grp_3='TMDE')

        # Hypomethylation  | No Change | UP         | mRNA decrease (TPDS)         | Protein increase (TMDE)
        self.get_grp(meth_c='neg', rna_c='-', prot_c='pos', grp_id='TMDE', reg_grp_1='TPDS_TMDE', reg_grp_3='TMDE')

        # Hypomethylation  | No Change | DOWN       | mRNA decrease (TPDS)         | Protein decrease (TMDS)
        self.get_grp(meth_c='neg', rna_c='-', prot_c='neg', grp_id='TMDS', reg_grp_1='TPDS_TMDS', reg_grp_3='TMDS')

        # Hypomethylation  | UP        | No Change  | Methylation decrease (MDE)   | Protein decrease (TMDS)
        self.get_grp(meth_c='neg', rna_c='pos', prot_c='-', grp_id='MDE_TMDS', reg_grp_1='MDE_TMDS', reg_grp_3='TMDS')

        # Hypomethylation  | DOWN      | No Change  | mRNA decrease (TPDS)         | Protein increase (TMDE)
        self.get_grp(meth_c='neg', rna_c='neg', prot_c='-', grp_id='TPDS_TMDE', reg_grp_1='TPDS_TMDE', reg_grp_3='TMDE')

        # -------------- Clusters with change in direction are mRNA and no DNA Methylation
        # No Change        | DOWN      | UP         | mRNA decrease (TPDS)         | Protein increase (TMDE)
        self.get_grp(meth_c='-', rna_c='neg', prot_c='pos', grp_id='TMDE', reg_grp_1='TPDS_TMDE', reg_grp_3='TMDE')

        # No Change        | UP        | DOWN       | mRNA increase (TPDE)         | Protein decrease (TMDS)
        self.get_grp(meth_c='-', rna_c='pos', prot_c='neg', grp_id='TMDS', reg_grp_1='TPDE_TMDS', reg_grp_3='TMDS')

        # -------------- Clusters with no methylation
        # No Change        | DOWN      | DOWN       | mRNA decrease (TPDS)         | None
        self.get_grp(meth_c='-', rna_c='neg', prot_c='neg', grp_id='TPDS',  reg_grp_1='TPDS', reg_grp_3='TPDS')

        # No Change        | UP        | UP         | mRNA increase (TPDE)         | None
        self.get_grp(meth_c='-', rna_c='pos', prot_c='pos', grp_id='TPDE',  reg_grp_1='TPDE', reg_grp_3='TPDE')

        # -------------- Clusters with only proteomics
        # No Change        | No Change | UP         | Protein increase (TMDE)      | None
        self.get_grp(meth_c='-', rna_c='-', prot_c='pos', grp_id='TMDE', reg_grp_1='TMDE', reg_grp_3='TMDE')

        # No Change        | No Change | DOWN       | Protein decrease (TMDS)      | None
        self.get_grp(meth_c='-', rna_c='-', prot_c='neg', grp_id='TMDS', reg_grp_1='TMDS', reg_grp_3='TMDS')

        # -------------- Clusters with only mRNA
        # No Change        | UP        | No Change  | mRNA increase (TPDE)         | Protein decrease (TMDS)
        self.get_grp(meth_c='-', rna_c='pos', prot_c='-', grp_id='TPDE_TMDS', reg_grp_1='TPDE_TMDS', reg_grp_3='TMDS')
        # No Change        | DOWN      | No Change  | mRNA decrease (TPDS)         | Protein increase (TMDE) | TPDS+TMDE
        self.get_grp(meth_c='-', rna_c='neg', prot_c='-', grp_id='TPDS_TMDE', reg_grp_1='TPDS_TMDE', reg_grp_3='TMDE')

        # -------------- Non-Coding gene clusters
        # Hypermethylation | No Change | No Change  | Methylation increase (ncRNA) | None
        self.get_grp(meth_c='pos', rna_c='-', prot_c='-', grp_id='MDS-ncRNA', reg_grp_1='MDS-ncRNA',
                     reg_grp_3='MDS-ncRNA', filter_list=self.non_coding_genes)
        # Hypomethylation  | No Change | No Change  | Methylation decrease (ncRNA) | None
        self.get_grp(meth_c='neg', rna_c='-', prot_c='-', grp_id='MDE-ncRNA', reg_grp_1='MDE-ncRNA',
                     reg_grp_3='MDE-ncRNA', filter_list=self.non_coding_genes)

        # Close the logfile
        if self.logfile is not None:
            self.logfile.close()
        return None

    def get_df(self):
        return self.df

    def integrate(self):
        # Here we combine the p-values and use a VAE to compute an integrated rank.
        combined_padj = []
        for i, p in enumerate(encoded_rcm['Protein_padj'].values):
            rna_p = encoded_rcm['RNA_padj'].values[i]
            cpg_p = encoded_rcm['CpG_padj'].values[i]
            combined_padj.append(combine_pvalues([p, rna_p, cpg_p])[1])
        encoded_rcm['combined_padj'] = combined_padj

    def get_grp(self, meth_c: str, rna_c: str, prot_c: str, grp_id: str, filter_list=None, reg_grp_1='', reg_grp_3=''):
        """ Compute all genes meeting a specific regulatory condition. """
        meth_change = self.df[self.meth_diff].values
        rna_change = self.df[self.rna_logfc].values
        prot_change = self.df[self.prot_logfc].values

        meth_padj = self.df[self.meth_padj].values
        rna_padj = self.df[self.rna_padj].values
        prot_padj = self.df[self.prot_padj].values

        grp = np.ones(len(meth_change))
        if rna_c == 'pos':
            grp *= 1.0 * (rna_change >= self.rna_logfc_cutoff) * (rna_padj <= self.rna_padj_cutoff)
        elif rna_c == 'neg':
            grp *= 1.0 * (rna_change <= (-1 * self.rna_logfc_cutoff)) * (rna_padj <= self.rna_padj_cutoff)
        else:
            # i.e. this gene should belong in the group if it fails to get assigned to the others i.e.
            # misses out based on the pvalue OR the logFC (hence the plus)
            grp *= (1.0 * (rna_padj > self.rna_padj_cutoff) + (1.0 * (abs(rna_change) < self.rna_logfc_cutoff)))

        if prot_c == 'pos':
            grp *= 1.0 * (prot_change >= self.prot_logfc_cutoff) * (1 * prot_padj <= self.prot_padj_cutoff)
        elif prot_c == 'neg':
            grp *= 1.0 * (prot_change <= (-1 * self.prot_logfc_cutoff)) * (1 * prot_padj <= self.prot_padj_cutoff)
        else:
            grp *= 1.0 * (1.0 * prot_padj > self.prot_padj_cutoff) + (abs(prot_change) < self.prot_logfc_cutoff)

        if meth_c == 'pos':
            grp *= 1.0 * (meth_change >= self.meth_diff_cutoff) * (meth_padj <= self.meth_padj_cutoff)
        elif meth_c == 'neg':
            grp *= 1.0 * (meth_change <= (-1 * self.meth_diff_cutoff)) * (meth_padj <= self.meth_padj_cutoff)
        else:
            grp *= 1.0 * ((1.0 * meth_padj > self.meth_padj_cutoff) + (abs(meth_change) < self.meth_diff_cutoff))

        # If we have a filter list of genes (e.g. for non-coding RNAs), then the genes MUST also be within this list
        genes_in_list = []
        if filter_list:
            for g in self.df[self.gene_id].values:
                if g in filter_list:
                    genes_in_list.append(1)
                else:
                    genes_in_list.append(0)
            grp *= np.array(genes_in_list)

        # Keep only the genes in this group
        grp_genes = self.df[self.gene_id].values[np.where(grp > 0)[0]]

        if self.debug:
            self.u.dp([grp_id, f'{meth_c} METH', f'{rna_c} RNA', f'{prot_c} PROT',
                  len(list(grp)), '\n', ', '.join(list(grp_genes))])
        if self.logfile is not None:
            # Print to logfile the results
            self.logfile.write(f'{meth_c},{rna_c},{prot_c},{grp_id},{len(list(grp_genes))}\n')
        # Add in the labels
        grp_ids = np.where(grp > 0)[0]

        # Create groups
        grp_1_labels = list(self.df[self.reg_grp_1_lbl].values)
        grp_2_labels = list(self.df[self.reg_grp_2_lbl].values)
        grp_3_labels = list(self.df[self.reg_grp_3_lbl].values)

        # Fill out the group labels 1 by 1 so that we can ensure that what we've added is correct.
        # Also if the user has set there to be overlaps (i.e. groups are not mutually exclusive) then this will print
        # out that there are possible duplicates.
        for i, g in enumerate(self.df[self.gene_id].values):
            if g in grp_genes:
                if grp_1_labels[i] and grp_1_labels[i] != 'None':
                    self.u.warn_p(["Possible duplicate Grouping 1:", grp_1_labels[i], reg_grp_1, g])
                    print(self.df[self.df[self.gene_id] == g])
                if grp_2_labels[i] and grp_2_labels[i] != 'None':
                    self.u.warn_p(["Possible duplicate Grouping 2:", grp_2_labels[i], grp_id, g])
                    print(self.df[self.df[self.gene_id] == g])
                if grp_1_labels[i] and grp_1_labels[i] != 'None':
                    self.u.warn_p(["Possible duplicate Grouping 3:", grp_3_labels[i], reg_grp_3, g])
                    print(self.df[self.df[self.gene_id] == g])
                grp_1_labels[i] = f'{reg_grp_1}_M-{meth_c}_R-{rna_c}_P-{prot_c}'
                grp_2_labels[i] = grp_id
                grp_3_labels[i] = reg_grp_3

        self.df[self.reg_grp_1_lbl] = grp_1_labels
        self.df[self.reg_grp_2_lbl] = grp_2_labels
        self.df[self.reg_grp_3_lbl] = grp_3_labels

        return self.df[self.gene_id][grp_ids]

    def save_df(self, output_filename):
        """ Save DF to a csv (mainly for R) """
        self.df.to_csv(output_filename, index=False)

    def get_genes_in_reg_grp(self, label, gene_label=None, reg_label=None):
        """ Get genes for a particular regulatory group. """
        reg_label = reg_label if reg_label else self.main_reg_label
        gene_label = self.gene_id if not gene_label else gene_label
        return list(set(self.df[self.df[reg_label] == label][gene_label].values))

    def get_all_assigned_genes(self, gene_label=None, reg_label=None):
        """ Get all genes assigned to a regulatory group (i.e. any that are not none) """
        reg_label = reg_label if reg_label else self.main_reg_label
        gene_label = self.gene_id if not gene_label else gene_label
        return list(set(self.df[self.df[reg_label] != 'None'][gene_label].values))

    def get_all_unassigned_genes(self, gene_label=None, reg_label=None):
        """ Get all genes that weren't assigned to a regulatory label. """
        gene_label = self.gene_id if not gene_label else gene_label
        reg_label = reg_label if reg_label else self.main_reg_label
        return list(set(self.df[self.df[reg_label] == 'None'][gene_label].values))


def filter_methylation_data_by_genes(cpg_df, gene_id, p_val, logfc):
    cpg_df_grped = cpg_df.groupby(gene_id)
    rows = []
    num_cpgs = []
    for cpg in cpg_df_grped:
        cpg = cpg[1]
        cpg = cpg[cpg[p_val] < 0.05]
        num_cpgs.append(len(cpg))
        if len(cpg) > 0:
            if len(cpg) < 3:
                add_row = True
            else:
                pos_cpg = cpg[cpg[logfc] > 0]
                neg_cpg = cpg[cpg[logfc] < 0]
                num_pos = len(pos_cpg)
                num_neg = len(neg_cpg)
                add_row = False
                if num_pos and num_pos/len(cpg) > 0.6:
                    cpg = pos_cpg
                    add_row = True
                elif num_neg and num_neg/len(cpg) > 0.6:
                    cpg = neg_cpg
                    add_row = True
            if add_row:
                max_cpg_idx = None
                max_t_value = 0  # absolute
                idxs = cpg.index
                for i, t in enumerate(cpg[logfc].values):
                    if abs(t) > abs(max_t_value):
                        max_t_value = t
                        max_cpg_idx = i
                rows.append(cpg[cpg.index == idxs[max_cpg_idx]].values[0])
    new_cpg_df = pd.DataFrame(rows, columns=cpg_df.columns)
    u = SciUtil()
    u.dp(['Originally had: ', len(cpg_df_grped), 'genes.\n', 'Filtered DF now has: ', len(new_cpg_df), ' genes.'])
    return new_cpg_df



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from sciviso import Emapplot
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from collections import defaultdict
from wordcloud import WordCloud


def plot_cluster_ORA(filename, gene_ratio='GeneRatio', count_column='Count', padj='p.adjust', overlap_column='geneID',
                     id_column='ID', label_column='Description', gene_ratio_min=0.05, padj_max=0.05, title='',
                     label_font_size=9, figsize=(3, 3), axis_font_size=6, min_count=20, max_count=200, min_overlap=4,
                     save_fig=True):
    """

    Parameters
    ----------
    filename
    gene_ratio
    count_column
    padj
    overlap_column
    id_column
    label_column
    gene_ratio_min
    padj_max
    title
    label_font_size
    figsize
    axis_font_size
    min_count
    max_count
    min_overlap
    save_fig

    Returns
    -------

    """
    df = pd.read_csv(f'{filename}')
    # Convert gene ratio to a number
    gr = df[gene_ratio].values
    gene_ratios = []
    for g in gr:
        g = g.split('/')
        g0 = float(g[0])
        g1 = float(g[1])
        gene_ratios.append(g0 / g1)
    df[gene_ratio] = gene_ratios
    df = df[df[gene_ratio] > gene_ratio_min]
    df = df[df[padj] < padj_max]
    if len(df) > 1:
        eplot = Emapplot(df, size_column=count_column, color_column=padj, id_column=id_column,
                         label_column=label_column, overlap_column=overlap_column, overlap_sep='/', title=title,
                         config={'figsize': figsize, 'label_font_size': label_font_size,
                                 'axis_font_size': axis_font_size})
        eplot.build_graph(min_overlap=min_overlap)
        plt.title(title, fontsize=18, fontweight='bold')
        plt.gca().set_clip_on = False

        if save_fig:
            plt.savefig(f'{filename.replace(".csv", "")}_Network.svg', bbox_inches='tight', transparent=True)
        plt.show()

        x, y = np.ogrid[:300, :300]
        plt.rcParams['svg.fonttype'] = 'none'  # Ensure text is saved as text
        plt.rcParams['figure.figsize'] = figsize
        mask = (x - 150) ** 2 + (y - 150) ** 2 > 130 ** 2
        mask = 255 * mask.astype(int)
        wordfeqs = defaultdict(int)
        for g in df[overlap_column].values:
            for w in g.split('/'):
                w = w.replace(' ', '.')
                wordfeqs[w] += 1
        total_words = len(wordfeqs)
        for w in wordfeqs:
            wordfeqs[w] = wordfeqs[w] / total_words
        # Compute the frequency of each word (since there are duplicates sometimes...)
        wordcloud = WordCloud(background_color="white", mask=mask, repeat=False).generate_from_frequencies(wordfeqs)
        wordcloud_svg = wordcloud.to_svg(embed_font=True)
        if save_fig:
            f = open(f'{filename.replace(".csv", "")}_WordCloud.svg', "w+")
            f.write(wordcloud_svg)
            f.close()
        plt.imshow(wordcloud, interpolation="bilinear")
        plt.axis("off")
        plt.show()


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from scivae import VAE, VAEStats
from scipy import stats
import numpy as np
import pandas as pd
from sciutil import SciUtil
import json


class RCMVAE:
    """
    Enables running a VAE on the RCM data.
    Takes in the output of rcm and then for each SiRCLe cluster, trains a VAE. It returns the dataset which can be
    used as input to calculating statistics on the DS.
    """

    def __init__(self, rcm_file: str,
                 meth_file: str, meth_sample_file: str,
                 rna_file: str, rna_sample_file: str,
                 protein_file: str, protein_sample_file: str,
                 output_folder: str, stat_condition_column: str,  # This is a a column that must be present in all sample dfs
                 sub_condition_id_column: str,
                 # For example stage of the patient.
                 config_file: str=None, regulatory_label='Regulation_Grouping_2',
                 run_name: str = None, normalise = 'columns'):
        self.config_json = f'{config_file}'
        self.feature_columns = ['RNA-LogFC',
                                'Protein-LogFC',
                                'CpG-LogFC',
                                'RNA-Tumor',
                                'RNA-Normal',
                                'Protein-Tumor',
                                'Protein-Normal']
        # Read in the data
        self.regulatory_label = regulatory_label
        self.rcm = pd.read_csv(rcm_file, index_col=0)
        self.meth_data = pd.read_csv(meth_file, index_col=0)
        self.rna_data = pd.read_csv(rna_file, index_col=0)
        self.protein_data = pd.read_csv(protein_file, index_col=0)
        self.run_name = run_name
        # Read in the sample dfs
        self.meth_sample_df = pd.read_csv(meth_sample_file)
        self.rna_sample_df = pd.read_csv(rna_sample_file)
        self.protein_sample_df = pd.read_csv(protein_sample_file)
        self.stat_condition_column = stat_condition_column
        self.stat_condition_id_column = sub_condition_id_column
        # For each of these files, normalise the rows to be between 0 and 1.
        self.output_folder = output_folder
        self.normalise = normalise
        self.sample_df = pd.DataFrame(columns=['case_id', 'condition_id', 'column_label', 'column_id', 'multi_loss'])
        self.u = SciUtil()

    def train_vae(self, config=None):
        """
        Sets up and normalises the data.
        :return:
        """
        for reg_label in self.rcm[self.regulatory_label].unique():
            if reg_label != "None":
                rcm_df = self.rcm[self.rcm[self.regulatory_label] == reg_label].copy()
                meth_data = self.compute_columns(self.align_to_rcm(self.meth_data, rcm_df), self.meth_sample_df,
                                                 'CpG', include_missing=False)
                rna_data = self.compute_columns(self.align_to_rcm(self.rna_data, rcm_df), self.rna_sample_df, 'RNA',
                                                include_missing=False)
                protein_data = self.compute_columns(self.align_to_rcm(self.protein_data, rcm_df,
                                                                      include_regulatory_label=True),
                                                    self.protein_sample_df, 'Protein', include_missing=False)
                r_df = pd.concat(
                    [protein_data,
                     rna_data,
                     meth_data], axis=1)
                # Create a sample df based on the columns
                train_df = self.build_training_df(r_df)
                self.train(train_df, self.feature_columns, reg_label, config)

    def train(self, train_df, feature_columns, reg_label, config):
        epochs = 100
        batch_size = 16
        if config is None:
            loss = {'loss_type': 'multi', 'distance_metric': 'mmd', 'mmd_weight': 0.1, 'multi_loss': ['mse', 'mse']}

            config = {"loss": loss,
                      "encoding": {"layers": [[{"num_nodes": 16, "activation_fn": "relu"},
                                               {"num_nodes": 16, "activation_fn": "relu"}]]},
                      "decoding": {"layers": [[{"num_nodes": 16, "activation_fn": "relu"},
                                               {"num_nodes": 16, "activation_fn": "relu"}]]},
                      "latent": {"num_nodes": 1},
                      "optimiser": {"params": {'learning_rate': 0.01}, "name": "adam"},
                      "input_size": [3, 4],
                      "output_size": [3, 4],
                      "epochs": epochs,
                      "batch_size": batch_size
            }
        else:
            epochs = config.get('epochs', epochs)
            batch_size = config.get('batch_size', batch_size)

        sub_sample = train_df
        if config.get('loss').get('loss_type') == 'multi':
            data_values = [sub_sample[['RNA-LogFC', 'Protein-LogFC', 'CpG-LogFC']].values,
                           sub_sample[['RNA-Tumor', 'RNA-Normal', 'Protein-Tumor', 'Protein-Normal']].values]
        else:
            data_values = sub_sample[feature_columns].values
        labels = list(sub_sample.index.values)
        vae_m = VAE(data_values, data_values, labels, config, vae_label=reg_label)
        # epochs=50, batch_size=50, train_percent=85.0, logging_dir=None, logfile=None
        vae_m.encode('default', train_percent=75.0, epochs=epochs, batch_size=batch_size,
                     logging_dir=self.output_folder,
                     logfile=f'VAE-logfile-{reg_label}-{self.run_name}.txt', early_stop=True)
        vae_m.save(weight_file_path=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-weights.h5',
                   optimizer_file_path=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-optimizer.json',
                   config_json=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-config.json')  # save the VAE
        vae_m.u.dp(["Saved VAE to current directory."])

    def get_decoding_for_cluster(self, reg_label, cases):
        weight_file_path = f'{self.output_folder}{reg_label}-{self.run_name}-VAE-weights.h5'
        optimizer_file_path = f'{self.output_folder}{reg_label}-{self.run_name}-VAE-optimizer.json'
        config_json = f'{self.output_folder}{reg_label}-{self.run_name}-VAE-config.json'
        rcm_df = self.rcm[self.rcm[self.regulatory_label] == reg_label].copy()
        meth_data = self.compute_columns(self.align_to_rcm(self.meth_data, rcm_df), self.meth_sample_df,
                                         'CpG', include_missing=False)
        rna_data = self.compute_columns(self.align_to_rcm(self.rna_data, rcm_df), self.rna_sample_df, 'RNA',
                                        include_missing=False)
        protein_data = self.compute_columns(self.align_to_rcm(self.protein_data, rcm_df,
                                                              include_regulatory_label=True),
                                            self.protein_sample_df, 'Protein', include_missing=False)
        r_df = pd.concat(
            [protein_data,
             rna_data,
             meth_data], axis=1)

        # Create a sample df based on the columns
        sample_df = self.build_stats_sample_df(r_df)
        train_df = self.build_training_df(r_df, False, cases)

        data_values = train_df[self.feature_columns].values
        labels = list(sample_df.index.values)
        config = {}
        with open(config_json, "r") as fp:
            config = json.load(fp)
        vae_m = VAE(data_values, data_values, labels, config, vae_label=reg_label)
        # Then decode the data
        vae_m.load(weight_file_path, optimizer_file_path, config_json)
        encoding = vae_m.encode_new_data(data_values, scale=False)
        decoding = vae_m.decoder.predict(encoding)
        enc_df = pd.DataFrame(data=encoding, columns=['VAE'])
        enc_df['id'] = list(train_df.index.values)
        dec_df = pd.DataFrame(data=decoding, columns=self.feature_columns)
        dec_df['id'] = list(train_df.index.values)
        train_df['id'] = list(train_df.index.values)
        return enc_df, dec_df, train_df

    def align_to_rcm(self, df, rcm_df, include_regulatory_label=False):
        aligned_df = pd.DataFrame()
        aligned_df['genes'] = rcm_df.index
        aligned_df.set_index('genes', inplace=True)
        aligned_df = aligned_df.join(df, how='left')  # Get all the data aligned to the same index
        aligned_df.fillna(0, inplace=True)  # Fill in the missing values with 0
        # normalise df
        normalised_aligned_df = self.normalise_df(aligned_df) # normalise the data
        if include_regulatory_label:
            normalised_aligned_df[self.regulatory_label] = rcm_df[self.regulatory_label].values

        return normalised_aligned_df

    def run_vae_stats(self, label='', include_missing=False, meth_sample_file = None, rna_sample_file = None,
                      protein_sample_file = None):
        # Enable a user to load in different sample files for different runs.
        if meth_sample_file is not None:
            self.meth_sample_df = pd.read_csv(meth_sample_file)
        if rna_sample_file is not None:
            self.rna_sample_df = pd.read_csv(rna_sample_file)
        if protein_sample_file is not None:
            self.protein_sample_df = pd.read_csv(protein_sample_file)

        # Now they all have the same index and are aligned so we can just build our training data by concating them
        for reg_label in self.rcm[self.regulatory_label].unique():
            if reg_label != "None":
                rcm_df = self.rcm[self.rcm[self.regulatory_label] == reg_label].copy()
                meth_data = self.compute_columns(self.align_to_rcm(self.meth_data, rcm_df), self.meth_sample_df,
                                                 'CpG', include_missing=False)
                rna_data = self.compute_columns(self.align_to_rcm(self.rna_data, rcm_df), self.rna_sample_df, 'RNA',
                                                include_missing=include_missing)
                protein_data = self.compute_columns(self.align_to_rcm(self.protein_data, rcm_df,
                                                                      include_regulatory_label=True),
                                                    self.protein_sample_df, 'Protein', include_missing=include_missing)
                r_df = pd.concat(
                    [protein_data,
                     rna_data,
                     meth_data], axis=1)

                # Create a sample df based on the columns
                sample_df = self.build_stats_sample_df(r_df)
                matched_cases = []
                for case in sample_df['case_id'].unique():
                    case_sample_df = sample_df[sample_df['case_id'] == case]
                    if len(case_sample_df) > 8:
                        matched_cases.append(case)
                    else:
                        continue  # print(case)
                sample_df = sample_df[sample_df['case_id'].isin(matched_cases)]
                self.u.err_p([reg_label])
                vs = VAEStats(r_df, sample_df, weight_file_path=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-weights.h5',
                              optimizer_file_path=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-optimizer.json',
                              config_json=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-config.json',
                              feature_columns=self.feature_columns)
                # Load config so we get the loss type
                fp = open(f'{self.output_folder}{reg_label}-{self.run_name}-VAE-config.json', 'r')
                config = json.load(fp)
                fp.close()
                if config.get('loss').get('loss_type') == 'multi':
                    stats_vals = vs.peform_DVAE_multiloss([['RNA-LogFC', 'Protein-LogFC', 'CpG-LogFC'],
                                                          ['RNA-Tumor', 'RNA-Normal', 'Protein-Tumor', 'Protein-Normal']],
                                                          column_to_align_to=['Protein-LogFC', 'RNA-LogFC', 'CpG-LogFC'])
                else:
                    stats_vals = vs.peform_DVAE(column_to_align_to=['Protein-LogFC', 'RNA-LogFC', 'CpG-LogFC'])

                vs.test_for_normality(stats_vals['base_mean_cond_0'])
                vs.test_for_normality(stats_vals['base_mean_cond_1'])
                stats_vals.to_csv(f'{self.output_folder}{reg_label}-{self.run_name}-stats{label}.csv')

    def build_stats_sample_df(self, df):
        sample_df = pd.DataFrame(columns=['case_id', 'condition_id', 'column_label', 'column_id', 'multi_loss'])
        # ['case_id', 'condition_id', 'column_label', 'column_id', 'multi_loss'] sample df columns
        case_ids = []
        condition_ids = []
        column_label = []
        column_id = []
        multi_loss = []
        for c in df.columns:
            if c != 'id' and c != self.regulatory_label:
                info = c.split('_')
                # # [f'{sub_condition}_{stats_cond_map}_{case}_{label}-Normal']
                case_ids.append(info[2])
                # ToDO Warn here
                condition_ids.append(int(info[1]))
                if 'LogFC' in c:
                    multi_loss.append(0)
                else:
                    multi_loss.append(1)
                column_id.append(c)
                column_label.append(info[-1])
        sample_df['case_id'] = case_ids
        sample_df['condition_id'] = condition_ids
        sample_df['column_label'] = column_label
        sample_df['column_id'] = column_id
        sample_df['multi_loss'] = multi_loss
        return sample_df

    def build_training_df(self, df, filter_extremes=True, selected_cases=None):
        # We want to add all the case data as training data and just keep the columns in the correct order
        # Basically we do this for all cases.
        cases = list(set(['_'.join(c.split('_')[:-1]) for c in df.columns if c != 'id']))  # ToDO: CHECK!!!!!
        train_df = pd.DataFrame()
        if selected_cases is not None:
            cases_subset = [c for c in cases if c.split('_')[-1] in selected_cases]
            cases = cases_subset
        for case in cases:
            case_cond_df = pd.DataFrame()
            case_cond_df['id'] = list(df.index.values)
            for col in self.feature_columns:
                try:
                    case_cond_df[col] = df[f'{case}_{col}'].values  # Get the column name from the case
                except:
                    continue  # self.u.warn_p([f'{case}_{col} not in df... Skipping...'])
            # Add this to the cond_1_sample_df
            if len(case_cond_df.columns) == len(self.feature_columns) + 1:  # For the index column
                train_df = train_df.append(case_cond_df)
        train_df.set_index('id', inplace=True)
        if filter_extremes:
            z_score = np.abs(stats.zscore(train_df, axis=1))
            max_z_score = np.max(z_score, axis=1)
            train_df = train_df[max_z_score < 2]
        return train_df

    def normalise_df(self, df):
        """
        Formats the dataframe.
        :return:
        """
        # Next normalise
        # We want to add all the case data as training data and just keep the columns in the correct order
        # Basically we do this for all cases.

        if self.normalise == 'rows':
            data = df.values.copy()
            # Min max scale this data
            data_values = []  # Basically just going to normalise each row
            for i, row in enumerate(data):
                if row.max() == 0:
                    data_values.append(row)
                else:
                    non_zero_min = np.min(row[row > 0])
                    non_zero_max = np.max(row[row > 0])
                    new_values = []
                    for j, val in enumerate(row):
                        if val > 0:
                            new_values.append((val - non_zero_min) / (non_zero_max - non_zero_min))
                        else:
                            new_values.append(0)
                    data_values.append(new_values)

            # Refill in the normalised protein data
            new_df = pd.DataFrame(data_values, index=df.index, columns=df.columns)
            return new_df
        elif self.normalise == 'columns':
            scaled_df = pd.DataFrame()
            scaled_df['genes'] = df.index.values
            numeric_cols = [c for c in df.columns if c != self.regulatory_label and c != 'id']
            # For each column, normalise to min-max but fist ommit any 0's
            for col in numeric_cols:
                values = df[col].values.copy()
                if values.max() == 0:
                    scaled_df[col] = values
                else:
                    non_zero_values = values[values != 0]
                    min_nz = np.min(non_zero_values)
                    max_nz = np.max(non_zero_values)
                    values[values != 0] = (values[values != 0] - min_nz) / (max_nz - min_nz)  # Min max scale and
                    # leave the rest 0's
                    scaled_df[col] = values
            scaled_df.set_index('genes', inplace=True)
            return scaled_df

    def compute_columns(self, df, sample_df, label, include_missing=False):
        # Leave does that had complete 0's as complete 0's
        # Next we want to build the features for the model (namely, the case wise change in each data type).
        vae_data_df = pd.DataFrame()
        missing_cond_0_data, missing_cond_1_data = 0, 0
        count_comparisons = 0

        vae_data_df['id'] = list(df.index.values)  # The index must be the gene ID
        # First split by the sub condition (we want to keep the data separate).
        # Get the stats condition map
        stats_cond_map = dict(zip(sample_df[self.stat_condition_column].values, sample_df[self.stat_condition_id_column].values))
        for sub_condition in sample_df[self.stat_condition_column].unique():
            # Get the mean of all normal samples
            sub_sample_df = sample_df[sample_df[self.stat_condition_column] == sub_condition]
            cond_0_columns_all = sub_sample_df[sub_sample_df['condition_id'] == 0]['column_id'].values
            # We use these if either value is missing.
            if len(cond_0_columns_all) == 0:
                continue  #print(sub_condition)
            cond_0_mean = np.mean(df[cond_0_columns_all].values, axis=1)

            for case in sub_sample_df['case_id'].unique():
                # Want to do condition 1 - condition 0 --> keeping with standard approach
                case_df = sub_sample_df[sub_sample_df['case_id'] == case]
                cond_0_columns = case_df[case_df['condition_id'] == 0]['column_id'].values
                cond_1_columns = case_df[case_df['condition_id'] == 1]['column_id'].values
                # Require that must have cond_1 at least
                if len(cond_1_columns) > 0:
                    if len(cond_0_columns) == 0 and (label == 'CpG' or include_missing):
                        missing_cond_0_data += 1
                        cond_1_values = np.mean(df[cond_1_columns].values, axis=1)
                        vae_data_df[
                            f'{sub_condition}_{stats_cond_map[sub_condition]}_{case}_{label}-Normal'] = cond_0_mean
                        vae_data_df[
                            f'{sub_condition}_{stats_cond_map[sub_condition]}_{case}_{label}-Tumor'] = cond_1_values
                        vae_data_df[
                            f'{sub_condition}_{stats_cond_map[sub_condition]}_{case}_{label}-LogFC'] = cond_1_values - cond_0_mean
                        count_comparisons += 1
                    elif len(cond_0_columns) == 0 and label != 'CpG':
                        continue  #print("missing...", label, case)
                    else:
                        cond_0_values = np.mean(df[cond_0_columns].values, axis=1)  # May only have 1 value but this enables reps
                        cond_1_values = np.mean(df[cond_1_columns].values, axis=1)
                        vae_data_df[f'{sub_condition}_{stats_cond_map[sub_condition]}_{case}_{label}-Normal'] = cond_0_values
                        vae_data_df[f'{sub_condition}_{stats_cond_map[sub_condition]}_{case}_{label}-Tumor'] = cond_1_values
                        vae_data_df[f'{sub_condition}_{stats_cond_map[sub_condition]}_{case}_{label}-LogFC'] = cond_1_values - cond_0_values
                        count_comparisons += 1
                    # Add a few new rows to the sample df
                else:
                    missing_cond_1_data += 1
        # Set index
        vae_data_df.set_index('id', inplace=True)
        return vae_data_df


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import numpy as np
import os
from scircm import SciRCM
from sciutil import SciException
import pandas as pd


class SciRCMException(SciException):
    def __init__(self, message=''):
        Exception.__init__(self, message)


class SciRCMnp(SciRCM):

    def __init__(self, meth_file: str, rna_file: str, rna_logfc: str, rna_padj: str, meth_diff: str, meth_padj: str,
                 gene_id: str, rna_padj_cutoff=0.05,
                 prot_padj_cutoff=0.05, meth_padj_cutoff=0.05, rna_logfc_cutoff=1.0, prot_logfc_cutoff=0.5,
                 meth_diff_cutoff=10, output_dir='.', output_filename=None, non_coding_genes=None, debug_on=False,
                 sep=',', bg_type='M|R', sciutil=None, logfile=None,
                 reg_grp_1_lbl='Regulation_Grouping_1',
                 reg_grp_2_lbl='Regulation_Grouping_2',
                 reg_grp_3_lbl='Regulation_Grouping_3',
                 main_reg_label='Regulation_Grouping_2'):
        super().__init__(meth_file, rna_file, None, rna_logfc, rna_padj, meth_diff, meth_padj, None, None,
                         gene_id,
                         rna_padj_cutoff, prot_padj_cutoff, meth_padj_cutoff,
                         rna_logfc_cutoff, prot_logfc_cutoff, meth_diff_cutoff, output_dir,
                         output_filename, non_coding_genes, debug_on, sep,
                         bg_type, sciutil, logfile, reg_grp_1_lbl=reg_grp_1_lbl,
                         reg_grp_2_lbl=reg_grp_2_lbl, reg_grp_3_lbl=reg_grp_3_lbl, main_reg_label=main_reg_label)
        self.bg_type = bg_type
        self.bg_list = ['M&R', 'M|R', '*']
        if bg_type not in self.bg_list:
            self.u.err_p(['ERROR: selected background type was not allowed, please choose from one of: ', self.bg_list,
                          '\n Note: | means OR and & means AND'])
        self.output_filename = output_filename if output_filename else f'scircm_r{rna_logfc_cutoff}-{rna_padj_cutoff}' \
                                                                       f'_m{meth_diff_cutoff}-{meth_padj_cutoff}.csv'
        self.meth_df = pd.read_csv(meth_file, sep=sep)
        self.rna_df = pd.read_csv(rna_file, sep=sep)
        self.non_coding_genes = non_coding_genes
        # Contains genes for the non-coding region (use for human only).
        self.df = None
        # Contains the vae data
        self.vae = None

    @staticmethod
    def _get_bg_filter(bg_type, rna_val, meth_val, meth_padj_cutoff, rna_padj_cutoff):
        c = 0
        if bg_type == 'M&R':  # Protein AND (DNA methylation OR RNA)
            c = 1 if rna_val <= rna_padj_cutoff and meth_val <= meth_padj_cutoff else 0
        elif bg_type == 'M|R':  # Protein OR methylation OR RNA
            c = 1 if rna_val <= rna_padj_cutoff else 0
            c += 1 if meth_val <= meth_padj_cutoff else 0
        elif bg_type == '*':  # Use all genes as the background
            c = 1
        return c

    def gen_bg(self, bg_type=None, rna_padj_cutoff=None, meth_padj_cutoff=None):
        """
        Generate a background dataset i.e. since the RCM requires at least two of the 3 datasets to
        have a p value beneath a threshold we reduce our dataset to be smaller.
        """
        bg_type = bg_type if bg_type else self.bg_type
        meth_padj_cutoff = meth_padj_cutoff if meth_padj_cutoff else self.meth_padj_cutoff
        rna_padj_cutoff = rna_padj_cutoff if rna_padj_cutoff else self.rna_padj_cutoff
        filter_vals = np.zeros(len(self.merged_df))
        meth_padj_values = self.merged_df[self.meth_padj].values

        # Choose the background dataframe
        for i, rna_padj in enumerate(self.merged_df[self.rna_padj].values):
            c = self._get_bg_filter(bg_type, rna_padj, meth_padj_values[i],
                                    meth_padj_cutoff, rna_padj_cutoff)
            filter_vals[i] = c

        df = self.merged_df.copy()
        # Filter the df to become the background df
        df['Number of significant datasets'] = filter_vals
        # Filter DF to only include those that are sig in two.
        df = df[df['Number of significant datasets'] >= 1]
        df = df.reset_index()  # Reset the index of the dataframe

        return df

    def merge_dfs(self, fill_protein=False, protein_cols=None):
        self.rna_df = self.rna_df.set_index(self.gene_id)
        self.df = self.rna_df.merge(self.meth_df, on=self.gene_id, how='outer', suffixes=['_r', '_m'])
        # Fill any of the values with NAs (p values need to be filled with 1's)
        self.df[[self.meth_padj]] = self.df[[self.meth_padj]].fillna(value=1.0)
        self.df[[self.rna_padj]] = self.df[[self.rna_padj]].fillna(value=1.0)
        # Fill the rest of the values with 0's
        self.df = self.df.fillna(0)
        self.df.to_csv(os.path.join(self.output_dir, "merged_df.csv"), index=False)
        self.merged_df = self.df.copy()

    def run_rcm(self):
        lbls = ['None'] * len(self.df)
        # Initialise the regulatory labels
        self.df[self.reg_grp_1_lbl] = lbls
        self.df[self.reg_grp_2_lbl] = lbls
        self.df[self.reg_grp_3_lbl] = lbls
        if self.logfile is not None:
            self.logfile.write(f'MethylationState,RNAseqState,Label,NumGenes\n')

        # -------------- Clusters with only mRNA
        # No Change        | UP        | No Change  | mRNA increase (TPDE)
        self.get_grp(meth_c='-', rna_c='pos', grp_id='TPDE', reg_grp_1='TPDE', reg_grp_3='TPDE')
        # No Change        | DOWN      | No Change  | mRNA decrease (TPDS)
        self.get_grp(meth_c='-', rna_c='neg', grp_id='TPDS', reg_grp_1='TPDS', reg_grp_3='TPDS')

        # -------------- Clusters with both mRNA and DNA Methylation
        # Hypomethylation  | UP        | No Change  | Methylation decrease (MDE)
        self.get_grp(meth_c='neg', rna_c='pos', grp_id='MDE', reg_grp_1='MDE', reg_grp_3='MDE')

        # Hypomethylation  | DOWN      | No Change  | mRNA decrease (TPDS)
        self.get_grp(meth_c='neg', rna_c='neg', grp_id='TPDS', reg_grp_1='TPDS', reg_grp_3='TPDS')

        # Hypermethylation  | UP        | No Change       | Methylation decrease (MDE)
        self.get_grp(meth_c='pos', rna_c='pos', grp_id='TPDE', reg_grp_1='TPDE', reg_grp_3='TPDE')

        # Hypermethylation | DOWN      | DOWN       | Methylation increase (MDS)   | None
        self.get_grp(meth_c='pos', rna_c='neg', grp_id='MDS', reg_grp_1='MDS', reg_grp_3='MDS')

        # Here we want to add Methylation driven genes
        # -------------- Non-Coding gene clusters
        # Hypermethylation | No Change | No Change  | Methylation increase (ncRNA) | None
        self.get_grp(meth_c='pos', rna_c='-', grp_id='MDS-ncRNA', reg_grp_1='MDS-ncRNA',
                     reg_grp_3='MDS-ncRNA', filter_list=self.non_coding_genes)
        # Hypomethylation  | No Change | No Change  | Methylation decrease (ncRNA) | None
        self.get_grp(meth_c='neg', rna_c='-', grp_id='MDE-ncRNA', reg_grp_1='MDE-ncRNA',
                     reg_grp_3='MDE-ncRNA', filter_list=self.non_coding_genes)

        # Close the logfile
        if self.logfile is not None:
            self.logfile.close()

    def get_df(self):
        return self.df

    def get_grp(self, meth_c, rna_c, grp_id, reg_grp_1, reg_grp_3, filter_list=None):
        """ Get a single group """
        meth_change = self.df[self.meth_diff].values
        rna_change = self.df[self.rna_logfc].values

        meth_padj = self.df[self.meth_padj].values
        rna_padj = self.df[self.rna_padj].values

        grp = np.ones(len(meth_change))
        if rna_c == 'pos':
            grp *= 1.0 * (rna_change >= self.rna_logfc_cutoff) * (rna_padj <= self.rna_padj_cutoff)
        elif rna_c == 'neg':
            grp *= 1.0 * (rna_change <= (-1 * self.rna_logfc_cutoff)) * (rna_padj <= self.rna_padj_cutoff)
        else:
            # i.e. this gene should belong in the group if it fails to get assigned to the others i.e.
            # misses out based on the pvalue OR the logFC (hence the plus)
            grp *= (1.0 * (rna_padj > self.rna_padj_cutoff) + (1.0 * (abs(rna_change) < self.rna_logfc_cutoff)))

        if meth_c == 'pos':
            grp *= 1.0 * (meth_change >= self.meth_diff_cutoff) * (meth_padj <= self.meth_padj_cutoff)
        elif meth_c == 'neg':
            grp *= 1.0 * (meth_change <= (-1 * self.meth_diff_cutoff)) * (meth_padj <= self.meth_padj_cutoff)
        else:
            grp *= 1.0 * ((1.0 * meth_padj > self.meth_padj_cutoff) + (abs(meth_change) < self.meth_diff_cutoff))

        # If we have a filter list of genes, then the genes MUST also be within this list
        genes_in_list = []
        if filter_list:
            for g in self.df[self.gene_id].values:
                if g in filter_list:
                    genes_in_list.append(1)
                else:
                    genes_in_list.append(0)
            grp *= np.array(genes_in_list)

        # Keep only the genes in this group
        grp_genes = self.df[self.gene_id].values[np.where(grp > 0)[0]]

        if self.debug:
            self.u.dp([grp_id, f'{meth_c} METH', f'{rna_c} RNA',
                  len(list(grp)), '\n', ', '.join(list(grp_genes))])
        if self.logfile is not None:
            # Print to logfile the results
            self.logfile.write(f'{meth_c},{rna_c},{grp_id},{len(list(grp_genes))}\n')
        # Add in the labels
        grp_ids = np.where(grp > 0)[0]

        # Create groups
        grp_1_labels = list(self.df[self.reg_grp_1_lbl].values)
        grp_2_labels = list(self.df[self.reg_grp_2_lbl].values)
        grp_3_labels = list(self.df[self.reg_grp_3_lbl].values)

        # Fill out the group labels 1 by 1 so that we can ensure that what we've added is correct.
        # Also if the user has set there to be overlaps (i.e. groups are not mutually exclusive) then this will print
        # out that there are possible duplicates.
        for i, g in enumerate(self.df[self.gene_id].values):
            if g in grp_genes:
                if grp_1_labels[i] and grp_1_labels[i] != 'None':
                    self.u.warn_p(["Possible duplicate Grouping 1:", grp_1_labels[i], reg_grp_1, g])
                    print(self.df[self.df[self.gene_id] == g])
                if grp_2_labels[i] and grp_2_labels[i] != 'None':
                    self.u.warn_p(["Possible duplicate Grouping 2:", grp_2_labels[i], grp_id, g])
                    print(self.df[self.df[self.gene_id] == g])
                if grp_1_labels[i] and grp_1_labels[i] != 'None':
                    self.u.warn_p(["Possible duplicate Grouping 3:", grp_3_labels[i], reg_grp_3, g])
                    print(self.df[self.df[self.gene_id] == g])
                grp_1_labels[i] = reg_grp_1
                grp_2_labels[i] = grp_id
                grp_3_labels[i] = reg_grp_3

        self.df[self.reg_grp_1_lbl] = grp_1_labels
        self.df[self.reg_grp_2_lbl] = grp_2_labels
        self.df[self.reg_grp_3_lbl] = grp_3_labels

        return self.df[self.gene_id][grp_ids]


###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################
import random

from statsmodels.stats.multitest import multipletests
from scivae import VAE
from scipy import stats
import numpy as np
import pandas as pd
from sciutil import SciUtil
import json


class RCMStats:
    """
    Enables running a VAE on the RCM data.

    Takes in the output of rcm and then for each SiRCLe cluster, trains a VAE. It returns the dataset which can be
    used as input to calculating statistics on the DS.
    """

    def __init__(self, rcm_file: str, patient_sample_file : str,
                 meth_file: str, meth_sample_file: str,
                 rna_file: str, rna_sample_file: str,
                 protein_file: str, protein_sample_file: str,
                 output_folder: str, condition_column: str,
                 column_id: str,
                 patient_id_column: str,
                 config_file: str = None,
                 regulatory_label='Regulation_Grouping_2',
                 run_name: str = None,
                 normalise='rows', verbose=False, missing_method='mean',
                 iid=False):
        self.config_json = f'{config_file}'
        self.feature_columns = ['RNA-LogFC',
                                'Protein-LogFC',
                                'CpG-LogFC',
                                'RNA-Tumor',
                                'RNA-Normal',
                                'Protein-Tumor',
                                'Protein-Normal']
        # Read in the data
        self.regulatory_label = regulatory_label
        self.rcm = pd.read_csv(rcm_file, index_col=0)
        self.meth_data = pd.read_csv(meth_file, index_col=0)
        self.rna_data = pd.read_csv(rna_file, index_col=0)
        self.protein_data = pd.read_csv(protein_file, index_col=0)
        self.run_name = run_name
        # Read in the sample dfs
        self.patient_id_column = patient_id_column  # This needs to be the same in all cases!
        self.patient_clinical_df = pd.read_csv(patient_sample_file)
        self.meth_sample_df = pd.read_csv(meth_sample_file)
        self.rna_sample_df = pd.read_csv(rna_sample_file)
        self.protein_sample_df = pd.read_csv(protein_sample_file)
        self.condition_column = condition_column or 'condition_id'
        self.column_id = column_id or 'column_id'
        self.missing_method = missing_method # either 'mean' or 'clinical'
        # For each of these files, normalise the rows to be between 0 and 1.
        self.output_folder = output_folder
        self.encoded_df = {}  # The encoded patient data.
        self.trained_vae = {}
        self.vae_input_df = {}
        self.train_df = {}
        self.raw_input_df = {} # The raw input data (i.e. no normaliseation).
        self.normalise = normalise
        self.u = SciUtil(debug_on=verbose)  # Set this optionally for verbosity
        self.u.warn_p(["WARNING: you cannot have underscores in your case_id names. Please check this before using"
                       "this tool."])
        self.check_files()
        self.iid = iid

    def save(self):
        """
        Save relevant files.

        Returns
        -------

        """
        encoded_df = pd.DataFrame()
        input_df = pd.DataFrame()
        raw_df = pd.DataFrame()
        for reg_label in self.encoded_df:
            df = self.encoded_df[reg_label]
            df[self.regulatory_label] = reg_label
            encoded_df = encoded_df.append(df)
            df = self.vae_input_df[reg_label]
            df[self.regulatory_label] = reg_label
            input_df = input_df.append(df)
            df = self.raw_input_df[reg_label]
            df[self.regulatory_label] = reg_label
            raw_df = raw_df.append(df)
        encoded_df.to_csv(f'{self.output_folder}encoded_df_{self.run_name}.csv')
        input_df.to_csv(f'{self.output_folder}vae_input_df_{self.run_name}.csv')
        raw_df.to_csv(f'{self.output_folder}raw_input_df_{self.run_name}.csv')

        # Also save the patient info
        self.patient_clinical_df.to_csv(f'{self.output_folder}patient_info_{self.run_name}.csv', index=False)

    def load_saved_vaes(self):
        for reg_label in self.rcm[self.regulatory_label].unique():
            weight_file_path = f'{self.output_folder}{reg_label}-{self.run_name}-VAE-weights.h5'
            optimizer_file_path = f'{self.output_folder}{reg_label}-{self.run_name}-VAE-optimizer.json'
            config_json = f'{self.output_folder}{reg_label}-{self.run_name}-VAE-config.json'
            with open(config_json, "r") as fp:
                config = json.load(fp)
            vae_m = VAE(np.ones((20, len(self.feature_columns))), np.ones((20, len(self.feature_columns))),
                        list(np.ones(20)), config, vae_label=reg_label)
            # Then decode the data
            vae_m.load(weight_file_path, optimizer_file_path, config_json)
            self.trained_vae[reg_label] = vae_m

    def load_saved_inputs(self, filename):
        """
        Optionally load a saved version of the input/training data patient data. Expected to have "id" as the
        first column, then the cases and values, and the regulatory label for that gene.
        Parameters
        ----------
        filename

        Returns
        -------

        """
        input_df = pd.read_csv(filename, index_col=0)
        for reg_label in self.rcm[self.regulatory_label].unique():
            self.vae_input_df[reg_label] = input_df[input_df[self.regulatory_label] == reg_label]

    def load_saved_raws(self, filename):
        """
        Optionally load a saved version of the input/training data patient data. Expected to have "id" as the
        first column, then the cases and values, and the regulatory label for that gene.
        Parameters
        ----------
        filename

        Returns
        -------

        """
        input_df = pd.read_csv(filename, index_col=0)
        for reg_label in self.rcm[self.regulatory_label].unique():
            self.raw_input_df[reg_label] = input_df[input_df[self.regulatory_label] == reg_label]

    def load_saved_encodings(self, filename):
        """
        Optionally load a saved version of the encoded patient data. Expected to have "id" as the
        first column, then the cases and values, and the regulatory label for that gene.
        Parameters
        ----------
        filename

        Returns
        -------

        """
        input_df = pd.read_csv(filename, index_col=0)
        for reg_label in self.rcm[self.regulatory_label].unique():
            self.encoded_df[reg_label] = input_df[input_df[self.regulatory_label] == reg_label]

    def check_files(self):
        """
        We need to check the loaded files to make sure they are in the correct format.

        Checks:
            1. The index of each of the data files overlaps with the regulatory clusters.
            2. Each sample file must have a) condition_column (default = 'condition_id') which has values 1 = tumour or
            0 = Normal and b) a column ID field (default = 'column_id') that maps to the column of that dataset,
            c) a case ID (that is a patient ID: default='case_id' that maps to a patient ID in the patient_clinical_df
            which has clinical information about the patient. case_ids can't have underscores in them.
        Returns
        -------

        """
        # Check 1.
        rcm_genes = set(self.rcm.index.values)
        protein_genes = set(self.protein_data.index.values)
        rna_genes = set(self.rna_data.index.values)
        meth_genes = set(self.meth_data.index.values)
        total_overlap = len(rcm_genes & rna_genes & protein_genes & meth_genes)
        self.u.dp(['Overlap between SiRCLe gene clusters and protein, RNA, and CpG: ', total_overlap, '\n',
                   f'If you used SiRCLe, this should be the total number of genes in your SiRCLe dataset. '
                   f'{len(rcm_genes)}', '\n First 5 gene IDs in RCM: ', list(rcm_genes)[0:5],
                   '\n First 5 gene IDs in Protein: ', list(protein_genes)[0:5],
                   '\n First 5 gene IDs in RNA: ', list(rna_genes)[0:5],
                   '\n First 5 gene IDs in Methylation: ', list(meth_genes)[0:5],
                   ])
        if total_overlap == 0:
            self.u.err_p(['You had no overlap between your gene identifiers! That is not good, nothing will work.'
                          'Please make sure your gene IDs are matching and are in the first column of your CSV file.'
                          '\nAlso ensure you are using a CSV not a TSV.'])
            return

        # Part 2.
        if not self.check_cols_exist():
            return
        self.build_sample_df()

    def get_sample_column(self, sample_df, case_id, condition_id):
        """
        Get the column identifier for a given case, sample df and condition ID.

        Parameters
        ----------
        sample_df
        case_id
        condition_id

        Returns
        -------

        """
        case_sample_df = sample_df[sample_df[self.patient_id_column] == case_id]
        if len(sample_df) > 0:
            value = case_sample_df[case_sample_df[self.condition_column] == condition_id][self.column_id].values
            if len(value) > 0:
                # only ever take one
                if len(value) > 1:
                    self.u.warn_p(['Had multiple samples for: ', case_id, 'just took the first one.',
                                   value])
                return value[0]
        return None

    def build_sample_df(self):
        """
        Builds a sample DF containing the columns that refer to the tumour and normal samples for the dataframes.
        This will help build the comparisons later on and also will make it clear which patients have had missing
        data added in.

        Returns
        -------

        """
        # Part 3. building the patient dataframe and making sure we have good data there (i.e. matching patients).
        protein_tumour = []
        rna_tumour = []
        meth_tumour = []
        protein_normal = []
        rna_normal = []
        meth_normal = []
        counts = []
        for i, case_id in enumerate(self.patient_clinical_df[self.patient_id_column].values):
            protein_normal.append(self.get_sample_column(self.protein_sample_df, case_id, 0))
            protein_tumour.append(self.get_sample_column(self.protein_sample_df, case_id, 1))
            rna_normal.append(self.get_sample_column(self.rna_sample_df, case_id, 0))
            rna_tumour.append(self.get_sample_column(self.rna_sample_df, case_id, 1))
            meth_normal.append(self.get_sample_column(self.meth_sample_df, case_id, 0))
            meth_tumour.append(self.get_sample_column(self.meth_sample_df, case_id, 1))
            # Add the number of non NAs we got
            count_data = 1 if protein_normal[i] is not None else 0
            count_data += 1 if protein_tumour[i] is not None else 0
            count_data += 1 if rna_normal[i] is not None else 0
            count_data += 1 if rna_tumour[i] is not None else 0
            count_data += 1 if meth_normal[i] is not None else 0
            count_data += 1 if meth_tumour[i] is not None else 0
            counts.append(count_data)
        df = self.patient_clinical_df.copy()
        # Make a new sample df incorperating this info.
        df['Protein Tumour'] = protein_tumour
        df['Protein Normal'] = protein_normal
        df['RNA Tumour'] = rna_tumour
        df['RNA Normal'] = rna_normal
        df['CpG Tumour'] = meth_tumour
        df['CpG Normal'] = meth_normal
        df['Sample counts'] = counts
        self.patient_clinical_df = df  # update possibly this is a bad idea...
        return df

    def check_cols_exist(self):
        """
        Check required columns exist in the sample data frames.
        Required columns:
        Each sample file must have a) condition_column (default = 'condition_id') which has values 1 = tumour or
            0 = Normal and b) a column ID field (default = 'column_id') that maps to the column of that dataset,
            c) a case ID (that is a patient ID: default='case_id' that maps to a patient ID in the patient_clinical_df
            which has clinical information about the patient. case_ids can't have underscores in them.
        Returns
        -------

        """
        cols = [self.condition_column, self.column_id, self.patient_id_column]
        # Check 2.
        for required_column in cols:
            if required_column not in self.meth_sample_df.columns:
                self.u.err_p([f'Your {required_column} was not in your methylation sample file?',
                              'This is needed. Nothing will work. Make sure your sample file is a CSV file.\n',
                              'Columns in the file you passed: ', self.meth_sample_df.columns])
                return False
            if required_column not in self.rna_sample_df.columns:
                self.u.err_p([f'Your {required_column} was not in your RNA sample file?',
                              'This is needed. Nothing will work. Make sure your sample file is a CSV file.\n',
                              'Columns in the file you passed: ', self.rna_sample_df.columns])
                return False

            if required_column not in self.protein_sample_df.columns:
                self.u.err_p([f'Your {required_column} was not in your Protein sample file?',
                              'This is needed. Nothing will work. Make sure your sample file is a CSV file.\n',
                              'Columns in the file you passed: ', self.protein_sample_df.columns])
                return False
        # Lastly check that the column id actually has overlaps with their data and they haven't done something dumb.
        protein_cols = set(self.protein_data.columns) & set(self.protein_sample_df[self.column_id].values)
        rna_cols = set(self.rna_data.columns) & set(self.rna_sample_df[self.column_id].values)
        methylation_cols = set(self.meth_data.columns) & set(self.meth_sample_df[self.column_id].values)
        if len(protein_cols) == 0:
            self.u.err_p([f'You made a mistake...  {self.column_id} does not actually map to your protein data file?',
                          'columns in your protein data file:', self.protein_data.columns,
                          '\n vs columns in your protein sample file: ', self.protein_sample_df[self.column_id].values])
            return False
        if len(rna_cols) == 0:
            self.u.err_p([f'You made a mistake...  {self.column_id} does not actually map to your protein data file?',
                          'columns in your protein data file:', self.rna_data.columns,
                          '\n vs columns in your protein sample file: ', self.rna_sample_df[self.column_id].values])
            return False

        if len(methylation_cols) == 0:
            self.u.err_p([f'You made a mistake...  {self.column_id} does not actually map to your protein data file?',
                          'columns in your protein data file:', self.meth_data.columns,
                          '\n vs columns in your protein sample file: ', self.meth_sample_df[self.column_id].values])
            return False

        # Final, check that there were overlapping patients in all and in the patient dataframe.
        patients_overlap = set(self.patient_clinical_df[self.patient_id_column].values) & \
                           set(self.protein_sample_df[self.patient_id_column].values) & \
                           set(self.rna_sample_df[self.patient_id_column].values) & \
                           set(self.meth_sample_df[self.patient_id_column].values)
        if len(patients_overlap) == 0:
            self.u.err_p([f'Hmmm your patient ids: ', self.patient_id_column, 'did not map in one of your sample files',
                          ' or the patient clinical info, please check them and then try again.'])
            return False

        # Sanitise the columns (replace '_' with .)
        self.patient_clinical_df[self.patient_id_column] = [c.replace('_', '.') for c in self.patient_clinical_df[self.patient_id_column].values]
        self.protein_sample_df[self.patient_id_column] = [c.replace('_', '.') for c in self.protein_sample_df[self.patient_id_column].values]
        self.rna_sample_df[self.patient_id_column] = [c.replace('_', '.') for c in self.rna_sample_df[self.patient_id_column].values]
        self.meth_sample_df[self.patient_id_column] = [c.replace('_', '.') for c in self.meth_sample_df[self.patient_id_column].values]

        # Yay they did good.
        return True

    def run_vae_stats(self, cond_label: str, cond0: str, cond1: str, label='', selected_cases=None):
        """
        Run stats comparing samples with condition 1 vs condition 0, the cond_label column. This is a column
        that must be present in all the sample data frames (for example, "gender" or "stage).

        Parameters
        ----------
        cond_label: the label of column in the patient sample df of
            the condition that we want to do that stats on e.g. gender, or stage
        cond0: the value that we want to be the control (e.g. stage 1)
        cond1: the value we want to test deviates from the control (e.g. stage 4)
        label: the label for this test
        include_missing: whether you want to include patients that are missing 1 or more data values e.g. only containes
            tumour for the protein data as opposed to both tumour and normal.
            If this is ticked, the data is filled with the mean value for that condition.

        Returns
        -------

        """
        cond1_cases_all = self.patient_clinical_df[self.patient_clinical_df[cond_label] == cond1][self.patient_id_column].values
        cond0_cases_all = self.patient_clinical_df[self.patient_clinical_df[cond_label] == cond0][self.patient_id_column].values

        if selected_cases is not None:
            cond1_cases_all = [c for c in cond1_cases_all if c in selected_cases]
            cond0_cases_all = [c for c in cond0_cases_all if c in selected_cases]

        # Note since not all of these values may have been included, keep only the cases that also were in the input df
        all_stats = pd.DataFrame()
        for reg_label in self.rcm[self.regulatory_label].unique():
            encoded_data = self.encoded_df[reg_label]  # Get the pre-encoded data for these patients...

            # Get it for each of the columns to align to
            cols_to_align = ['Protein-LogFC', 'RNA-LogFC', 'CpG-LogFC']
            cond0_cases = list(set([col for col in encoded_data.columns if col in cond0_cases_all]))
            cond1_cases = list(set([col for col in encoded_data.columns if col in cond1_cases_all]))

            alignment_column_1_values = []
            alignment_column_0_values = []
            # Also we want to make sure the columns are aligned to something biologicallly meaningful, so we want
            # to add in their "input data" so it can be aligned to this and we do this on the non-normalised data
            raw_input_df = self.raw_input_df[reg_label]
            for c in cols_to_align:
                # Get the mean value for this condition
                cols = [col for col in raw_input_df.columns if c in col and col.split('_')[0] in cond1_cases]
                data = np.nanmean(raw_input_df[cols].values, axis=1)
                alignment_column_1_values.append(data)
                cols = [col for col in raw_input_df.columns if c in col and col.split('_')[0] in cond0_cases]
                data = np.nanmean(raw_input_df[cols].values, axis=1)
                alignment_column_0_values.append(data)

            # cond_0_encodings = {case_id: [encoded_data[case_id].values] for case_id in cond0_cases}
            # cond_1_encodings = {case_id: [encoded_data[case_id].values] for case_id in cond1_cases}
            stats_df = self.make_stats_df(test_type='mannwhitneyu', id_vals=encoded_data['id'].values,
                                          cond_1_encodings=encoded_data[cond1_cases],
                                          cond_0_encodings=encoded_data[cond0_cases],
                                          column_to_align_to=cols_to_align,
                                          alignment_column_1_values=alignment_column_1_values,
                                          alignment_column_0_values=alignment_column_0_values, cond0=cond0, cond1=cond1)
            stats_df[self.regulatory_label] = reg_label
            # Save the averages from the cols to align to as well
            for i, c in enumerate(cols_to_align):
                stats_df[f'{c} mean ({cond1})'] = alignment_column_1_values[i]
                stats_df[f'{c} mean ({cond0})'] = alignment_column_0_values[i]
                stats_df[f'{c} mean ({cond1}-{cond0})'] = alignment_column_1_values[i] - alignment_column_0_values[i]

            self.test_for_normality(stats_df[f'Integrated mean ({cond0})'], f'{reg_label} Integrated mean ({cond0})')
            self.test_for_normality(stats_df[f'Integrated mean ({cond1})'], f'{reg_label} Integrated mean ({cond1})')
            all_stats = all_stats.append(stats_df)
        all_stats.to_csv(f'{self.output_folder}stats_{cond1}-{cond0}_{self.run_name + label}.csv')
        return all_stats

    def test_for_normality(self, values, label, test_type: str = "shapiro"):
        """ Perform a test for normality."""
        k2, p = stats.normaltest(values)
        if p < 0.05:  # null hypothesis: x comes from a normal distribution
            print(f'{label}: NOT normally distributed')
            return False
        return True

    def make_stats_df(self, test_type, id_vals, cond_1_encodings, cond_0_encodings, column_to_align_to,
                      alignment_column_1_values, alignment_column_0_values, cond0, cond1):
        # Now we want to perform the differential test on the data between cond 1 - cond 0
        # If we have multiple samples we need to do this for each one
        if len(id_vals) > 0:
            stat_vals = []
            p_vals = []
            base_means_cond_0 = []
            base_means_cond_1 = []
            num_cond_0 = 0
            num_cond_1 = 0

            # For each case in the encodings we want to collect the values
            for i in range(0, len(id_vals)):
                # ToDo: extend to anova or other statistical tests for more data types.
                cases_0_vals = cond_0_encodings.values[i]
                cases_1_vals = cond_1_encodings.values[i]
                num_cond_0 = len(cases_0_vals)
                num_cond_1 = len(cases_1_vals)
                # potentially wrap a try catch if there are all even numbers
                if test_type == 't-test':
                    t_stat, p_val = stats.ttest_ind(cases_1_vals, cases_0_vals)
                else:
                    t_stat, p_val = stats.mannwhitneyu(cases_1_vals, cases_0_vals)
                if p_val == 0 or p_val > 1:
                    p_val = 1.0
                stat_vals.append(t_stat)
                p_vals.append(p_val)
                base_mean_cond_1 = np.nanmean(cases_1_vals)
                base_mean_cond_0 = np.nanmean(cases_0_vals)
                base_means_cond_0.append(base_mean_cond_0)
                base_means_cond_1.append(base_mean_cond_1)
            # Now we have the p-values we can perform the correction
            reg, corrected_p_vals, a, b = multipletests(p_vals, method='fdr_bh', alpha=0.05, returnsorted=False)
            # Return something similar to what you'd get from DEseq2
            stats_df = pd.DataFrame()
            stats_df['id'] = id_vals
            stats_df[f'{test_type} stat ({cond1}-{cond0})'] = stat_vals
            stats_df[f'Integrated padj ({cond1}-{cond0})'] = corrected_p_vals
            stats_df[f'Integrated pval ({cond1}-{cond0})'] = p_vals
            # Check if we have a column to align to
            base_means_cond_1 = np.array(base_means_cond_1)
            base_means_cond_0 = np.array(base_means_cond_0)
            if column_to_align_to is not None:
                # Go through each one and stop if we get over 0.5 correlation
                for col_i in range(0, len(alignment_column_0_values)):
                    mean_col_0 = alignment_column_0_values[col_i]  # Across genes
                    mean_col_1 = alignment_column_1_values[col_i]
                    col_0_corr = np.corrcoef(mean_col_0, base_means_cond_0)[0, 1]
                    col_1_corr = np.corrcoef(mean_col_1, base_means_cond_1)[0, 1]
                    if abs(col_0_corr) > 0.5 or abs(col_1_corr) > 0.5:
                        if abs(col_0_corr) > abs(col_1_corr):
                            direction = -1 if col_0_corr < 0 else 1
                        else:
                            direction = -1 if col_1_corr < 0 else 1
                        # Convert both
                        base_means_cond_0 = direction * base_means_cond_0
                        base_means_cond_1 = direction * base_means_cond_1
                        break # If none of them meet it then we don't change anything
            # Compute difference as the distance between the two
            distances = []
            for i, cond_0 in enumerate(base_means_cond_0):
                if cond_0 < 0:
                    distances.append(base_means_cond_1[i] + abs(cond_0))
                else:
                    distances.append(base_means_cond_1[i] - abs(cond_0))
            stats_df[f'Integrated diff ({cond1}-{cond0})'] = distances
            stats_df[f'Integrated mean ({cond0})'] = base_means_cond_0
            stats_df[f'Integrated mean ({cond1})'] = base_means_cond_1
            self.u.dp(['Summary\n', f'Cond1: {num_cond_1} vs Cond0: {num_cond_0}\n',
                       stats_df.describe()])
            # Also make a copy that also contains all the info from all the cases
            # make this optional later on...
            for c in cond_0_encodings:
                stats_df[f'{cond0}_{c}'] = cond_0_encodings[c].values
            for c in cond_1_encodings:
                stats_df[f'{cond1}_{c}'] = cond_1_encodings[c].values
            return stats_df
        else:
            # Only one value so just do the test once.
            cases_0_vals = [c for c in cond_0_encodings.values]
            cases_1_vals = [c for c in cond_1_encodings.values]
            t_stat, p_val = stats.mannwhitneyu(cases_1_vals, cases_0_vals)
            return t_stat, p_val

    def merge_data(self, protein_data, rna_data, meth_data):
        return pd.concat([protein_data, rna_data, meth_data], axis=1)

    def train_vae(self, cases, config=None, include_missing=True):
        """

        Parameters
        ----------
        cases: a list of patient identifiers which should be used for training, these are ideally, high quality with
        no missing data.
        config: a dictionary of configuration for the VAE training
        include_missing: whether or not to include patients that are missing some data (it is filled in by the mean value)
        note we only allow missing in terms of the "normal" not the tumour.
        Returns
        -------

        """
        encoded_df = pd.DataFrame()
        for reg_label in self.rcm[self.regulatory_label].unique():
            if reg_label != "None":
                rcm_df = self.rcm[self.rcm[self.regulatory_label] == reg_label].copy()
                meth_data = self.compute_columns_training(self.align_to_rcm(self.meth_data, rcm_df), self.meth_sample_df, 'CpG',
                                                          include_missing)
                rna_data = self.compute_columns_training(self.align_to_rcm(self.rna_data,  rcm_df), self.rna_sample_df, 'RNA',
                                                         include_missing)
                protein_data = self.compute_columns_training(self.align_to_rcm(self.protein_data, rcm_df,
                                                                               include_regulatory_label=True),
                                                             self.protein_sample_df,
                                                             'Protein', include_missing)

                r_df = self.merge_data(protein_data, rna_data, meth_data)
                self.vae_input_df[reg_label] = r_df.copy()  # Keep track of the training dataframe for these patients
                # Also do the same for a non-normalised version, we want this so that we can actually get the change
                # in protein etc.
                nn_meth_data = self.compute_columns_training(self.align_to_rcm(self.meth_data, rcm_df, normalise=False),
                                                          self.meth_sample_df, 'CpG',
                                                          include_missing)
                nn_rna_data = self.compute_columns_training(self.align_to_rcm(self.rna_data, rcm_df, normalise=False), self.rna_sample_df,
                                                         'RNA',
                                                         include_missing)
                nn_protein_data = self.compute_columns_training(self.align_to_rcm(self.protein_data, rcm_df, normalise=False,
                                                                               include_regulatory_label=True),
                                                             self.protein_sample_df,
                                                             'Protein', include_missing)
                self.raw_input_df[reg_label] = self.merge_data(nn_protein_data, nn_rna_data, nn_meth_data)
                # Now we need to filter out patients that didn't have the required matching data.
                train_df = self.build_training_df(r_df, selected_cases=cases)
                self.train_df[reg_label] = train_df.copy()
                self.train(train_df, self.feature_columns, reg_label, config)
                # Encode data with trained VAE for all patients.
                # encoding has genes as the rows IDs and patients as the columns, so we're basically building it up
                # for each of the patients
                reg_encoded_df = self.get_encoding(r_df, reg_label)
                reg_encoded_df[self.regulatory_label] = reg_label
                # Keep track of this for quick access
                self.encoded_df[reg_label] = reg_encoded_df
                encoded_df = encoded_df.append(reg_encoded_df)
        # Keep track of the patient encodings.
        return encoded_df

    def build_training_df(self, df, filter_extremes=True, selected_cases=None):
        """
        Selected cases are the cases with matched tumour and normal for all conditions (or a selection
        of cases that are used for training). This should be set otherwise all cases will be used and that
        could be suboptimal.
        Parameters
        ----------
        df: dataframe
        filter_extremes: whether or not to remove data (genexpatient) if that value is > 2 out of z score true by default
        selected_cases: the cases that are going to be used for training (list)

        Returns
        -------

        """
        # We want to add all the case data as training data and just keep the columns in the correct order
        # Basically we do this for all cases.
        cases = self.patient_clinical_df[self.patient_id_column].values
        train_df = pd.DataFrame()
        case_ids = []
        included_cases = []

        if selected_cases is not None:
            cases_subset = [c for c in cases if c in selected_cases]
            cases = cases_subset
        case_genes = None
        if self.iid:
            # Here is for the person with lots of money and resources who could actually make a big enough dataset
            # Since I think the above works, maybe next time spend your money on better things like make a new journal
            # where it's free to publish or the scientists actually get the royalties or something ya know...
            # To do this, we take a single gene from each patient, i.e. first collect the patients, then subsample
            # a single row that is unique between.
            patients = cases #list(set([c.split('_')[0] for c in df.columns]))
            n_genes = len(patients)
            if n_genes > len(df.index.values) or self.iid > len(df.index.values):
                n_genes = len(df.index.values) - 1
            # Randomly select the a sample of genes for that patient
            case_genes = []
            for case_idx, case in enumerate(cases):
                genes_for_patient = random.sample(list(enumerate(df.index.values)), n_genes)
                case_genes.append([[c[0] for c in genes_for_patient], [c[1] for c in genes_for_patient]])
            #case_genes = random.sample(list(enumerate(df.index.values)), n_genes)
        for case_idx, case in enumerate(cases):
            try:
                case_cond_df = pd.DataFrame()
                if self.iid:
                    idval = case_genes[case_idx][1]
                    case_cond_df['id'] = idval
                else:
                    case_cond_df['id'] = list(df.index.values)
                for col in self.feature_columns:

                        if self.iid:
                            valval = case_genes[case_idx][0]
                            v = df[f'{case}_{col}'].values[valval]
                            # Just select the single gene that we're interested in
                            case_cond_df[col] = v
                        else:
                            case_cond_df[col] = df[f'{case}_{col}'].values  # Get the column name from the case

                # Add this to the cond_1_sample_df
                if len(case_cond_df.columns) == len(self.feature_columns) + 1:  # For the index column
                    train_df = train_df.append(case_cond_df)
                    # Add the length of this to the case_ids list so we can extract this patient's information later
                    case_ids += [case] * len(case_cond_df)
                    included_cases.append(case)
            except:
                continue
        train_df.set_index('id', inplace=True)
        if filter_extremes:
            z_score = np.abs(stats.zscore(train_df[self.feature_columns].values, axis=1))
            max_z_score = np.max(z_score, axis=1)
            train_df = train_df[max_z_score < 2]
        self.u.dp([f'{len(included_cases)} had matched data.', included_cases])
        return train_df

    def normalise_df(self, df):
        """
        Normalise the dataframe either at a row level or a column level.

        Parameters
        ----------
        df

        Returns
        -------

        """
        # Next normalise
        # We want to add all the case data as training data and just keep the columns in the correct order
        # Basically we do this for all cases.

        if self.normalise == 'rows':
            data = df.values.copy()
            # Min max scale this data
            data_values = []  # Basically just going to normalise each row
            for i, row in enumerate(data):
                if row.max() == 0:
                    data_values.append(row)
                else:
                    non_zero_min = np.min(row[row > 0])
                    non_zero_max = np.max(row[row > 0])
                    new_values = []
                    for j, val in enumerate(row):
                        if val > 0:
                            new_values.append((val - non_zero_min) / (non_zero_max - non_zero_min))
                        else:
                            new_values.append(0)
                    data_values.append(new_values)

            # Refill in the normalised protein data
            new_df = pd.DataFrame(data_values, index=df.index, columns=df.columns)
            return new_df
        elif self.normalise == 'columns':
            scaled_df = pd.DataFrame()
            scaled_df['genes'] = df.index.values
            numeric_cols = [c for c in df.columns if c != self.regulatory_label and c != 'id']
            # For each column, normalise to min-max but fist ommit any 0's
            for col in numeric_cols:
                values = df[col].values.copy()
                if values.max() == 0:
                    scaled_df[col] = values
                else:
                    non_zero_values = values[values != 0]
                    min_nz = np.min(non_zero_values)
                    max_nz = np.max(non_zero_values)
                    values[values != 0] = (values[values != 0] - min_nz) / (max_nz - min_nz)  # Min max scale and
                    # leave the rest 0's
                    scaled_df[col] = values
            scaled_df.set_index('genes', inplace=True)
            return scaled_df

    def fill_missing(self, df, case_id, sample_df, label):
        """
        If the user wants to include the misisng data, do it either by the mean, or by a series of the
        values associated with the patient sample types.

        Parameters
        ----------
        case_id

        Returns
        -------

        """
        if self.missing_method == 'mean':
            cond_0 = list(sample_df[sample_df[self.condition_column] == 0][self.column_id].values)
            cond_0_cols = [c for c in cond_0 if c in list(df.columns)]  # Ensure it is in the cols
            return np.mean(df[cond_0_cols].values, axis=1)
        if self.missing_method == 'clinical':
            # For now just do it on patient stage and age of patient
            case_info = self.patient_clinical_df[self.patient_clinical_df[self.patient_id_column] == case_id]
            # Get the information we care about ToDo: generalise
            age = case_info['AgeGrouped'].values[0]
            stage = case_info['Stage'].values[0]
            # Now let's get the values for that age and stage
            other_cases = self.patient_clinical_df[self.patient_clinical_df['AgeGrouped'] == age]
            other_cases = other_cases[other_cases['Stage'] == stage]
            # Now return the values for this for the given label that was asked for.
            cond_0_columns = [c for c in other_cases[f'{label} Normal'].values if c in df.columns]  # The normal values for the cases with same
            # age and stage
            return np.mean(df[cond_0_columns].values, axis=1)

    def compute_columns_training(self, df, sample_df, label, include_missing):
        """
        Goal is to add in the data for the columns for the VAE training. For this context we're interested
        in the tumour, normal, and logFC between the two.
        Parameters
        ----------
        df
        label
        include_missing

        Returns
        -------

        """
        vae_data_df = pd.DataFrame()  # This will be the data used for input
        vae_data_df['id'] = df.index.values
        # Get a baseline for all "normal"
        cond_0 = sample_df[sample_df[self.condition_column] == 0]  # i.e. normal
        cond_0_columns_all = [c for c in cond_0[self.column_id] if c is not None]
        if len(cond_0_columns_all) == 0:
            self.u.warn_p(['Dataset passed had no normal columns... you have an error please fix it!'])
            return
        # baseline.
        for case in self.patient_clinical_df[self.patient_id_column].unique():
            # Want to do condition 1 - condition 0 --> keeping with standard approach
            case_df = self.patient_clinical_df[self.patient_clinical_df[self.patient_id_column] == case]
            cond_0_column = case_df[f'{label} Normal'].values[0]
            cond_1_column = case_df[f'{label} Tumour'].values[0]
            # Require that must have cond_1 at least
            if cond_1_column is not None:
                if cond_0_column is None:
                    if include_missing:
                        cond_0_mean = self.fill_missing(df, case, sample_df, label)
                        # If it doesn't have any NA, include this otherwise ommit the data
                        if not np.isnan(cond_0_mean).any():
                            cond_1_values = df[cond_1_column].values
                            vae_data_df[f'{case}_{label}-Normal'] = cond_0_mean
                            vae_data_df[f'{case}_{label}-Tumor'] = cond_1_values
                            vae_data_df[f'{case}_{label}-LogFC'] = cond_1_values - cond_0_mean
                else:
                    # May only have 1 value for a patient - this summarises the replicates
                    cond_0_values = df[cond_0_column].values
                    cond_1_values = df[cond_1_column].values
                    vae_data_df[f'{case}_{label}-Normal'] = cond_0_values
                    vae_data_df[f'{case}_{label}-Tumor'] = cond_1_values
                    vae_data_df[f'{case}_{label}-LogFC'] = cond_1_values - cond_0_values
        # Set index
        vae_data_df.set_index('id', inplace=True)
        return vae_data_df

    def train(self, train_df, feature_columns, reg_label, config=None):
        """
        Train the vae.
        Parameters
        ----------
        train_df: a dataframe of rows which are patient gene values (i.e. n rows = patients x genes) and summarised
        columns (i.e. RNA logFC, Protein logFC, ... etc for that patient for that gene).
        feature_columns: the columns used for the VAE.
        reg_label: the SiRCle cluster.
        config: a dictionary of the config for the VAE. If None, then a default is used.

        Returns
        -------

        """
        if config is None:
            config = {"loss": {'loss_type': 'mse', 'distance_metric': 'mmd', 'mmd_weight': 0.25},
                      "encoding": {"layers": [{"num_nodes": 5, "activation_fn": "relu"}]},
                      "decoding": {"layers": [{"num_nodes": 5, "activation_fn": "relu"}]},
                      "latent": {"num_nodes": 1},
                      "optimiser": {"params": {'learning_rate': 0.01}, "name": "adam"},
                      "epochs": 200,
                      "batch_size": 16,
                      "scale_data": False
                      }
        epochs = config.get('epochs', 100)
        batch_size = config.get('batch_size', 16)
        data_values = train_df[feature_columns].values
        vae_m = VAE(data_values, data_values, list(train_df.index.values), config, vae_label=reg_label)
        vae_m.encode('default', train_percent=75.0, epochs=epochs, batch_size=batch_size,
                     logging_dir=self.output_folder,
                     logfile=f'VAE-logfile-{reg_label}-{self.run_name}.txt', early_stop=True)
        vae_m.save(weight_file_path=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-weights.h5',
                   optimizer_file_path=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-optimizer.json',
                   config_json=f'{self.output_folder}{reg_label}-{self.run_name}-VAE-config.json')  # save the VAE
        # Save an encoding of the data in the class and return it to the user.
        self.trained_vae[reg_label] = vae_m
        vae_m.u.dp(["Saved VAE to current directory."])
        return vae_m

    def get_decoding(self, reg_label):
        vae_m = self.trained_vae[reg_label]
        # Get the encoded data for that label
        encoding = self.encoded_df[reg_label]
        decoding_df = encoding[['id', self.regulatory_label]].copy()
        for column in encoding.columns:
            if column != 'id' and column != self.regulatory_label:
                encoding_vals = encoding[column].values
                decoding = vae_m.decoder.predict(np.array([np.array(c) for c in encoding_vals]))
                # The decoding will be in the same format as what we in
                for i, c in enumerate(self.feature_columns):
                    decoding_df[f'{column}_{c}'] = decoding[:, i]
        return decoding_df

    def get_encoding(self, input_df, reg_label):
        # Do this for each case, and then save as a DF for this cluster
        r_df = pd.DataFrame()
        r_df['id'] = input_df.index
        # Just make sure everything is in the right order
        vae_m = self.trained_vae[reg_label]
        for case_id in self.patient_clinical_df[self.patient_id_column].unique():
            case_cond_df = pd.DataFrame()
            for col in self.feature_columns:
                try:
                    case_cond_df[col] = input_df[f'{case_id}_{col}'].values  # Get the column name from the case
                except:
                    continue
            if len(case_cond_df.columns) == len(self.feature_columns):
                encoded_data = vae_m.encode_new_data(case_cond_df.values, scale=False)
                # For each of the cases add it as a column
                r_df[case_id] = encoded_data
            # else:
            #     self.u.dp(["CaseID: ", case_id, " must have been missing tumour data..."])
        self.encoded_df[reg_label] = r_df  # Save to state as well while we're going.
        return r_df

    def align_to_rcm(self, df, rcm_df, include_regulatory_label=False, normalise=True):
        aligned_df = pd.DataFrame()
        aligned_df['genes'] = rcm_df.index
        aligned_df.set_index('genes', inplace=True)
        aligned_df = aligned_df.join(df, how='left')  # Get all the data aligned to the same index
        aligned_df.fillna(0, inplace=True)  # Fill in the missing values with 0
        # normalise df
        if normalise:
            normalised_aligned_df = self.normalise_df(aligned_df) # normalise the data
            if include_regulatory_label:
                normalised_aligned_df[self.regulatory_label] = rcm_df[self.regulatory_label].values
            return normalised_aligned_df
        else:
            if include_regulatory_label:
                aligned_df[self.regulatory_label] = rcm_df[self.regulatory_label].values
                return aligned_df
            return aligned_df
